#pragma once

class AudioAppDemo : public AudioAppComponent
{
public:
    AudioAppDemo()
    {
        setAudioChannels (0, 2);

        createGUI();
        setSize (800, 600);

        midiInputChanged();
    }

    ~AudioAppDemo() override
    {
        disableAllMidiOutputs();
        disableAllMidiInputs();
        shutdownAudio();
    }


    void disableAllMidiInputs()
    {
        auto midiInputList = juce::MidiInput::getAvailableDevices();

        // Ensure all MIDI inputs are disabled
        for (auto& x : midiInputList)
            if (deviceManager.isMidiInputDeviceEnabled(x.identifier))
                deviceManager.setMidiInputDeviceEnabled(x.identifier, false);

        // Remove previously existing callback
        deviceManager.removeMidiInputDeviceCallback("", &midiInputCollector);
    }


    void midiInputChanged()
    {
        disableAllMidiInputs();

        int itemId = midiInputSelection.getSelectedId();
        jassert(itemId != 0);        // Ensure our test has a valid comboBox item selected
        int listIndex = itemId - 1;

        auto midiInputList = juce::MidiInput::getAvailableDevices();

        // Enable the requested MIDI input and get data from all enabled MIDI inputs
        deviceManager.setMidiInputDeviceEnabled(midiInputList[listIndex].identifier, true);
        deviceManager.addMidiInputDeviceCallback("", &midiInputCollector);
    }


    void disableAllMidiOutputs()
    {
        if (midiOutputDevice == nullptr)
            return;

        midiOutputDevice->clearAllPendingMessages();
        midiOutputDevice->stopBackgroundThread();
        midiOutputDevice.reset();
    }


    void midiOutputChanged()
    {
        disableAllMidiOutputs();

        int itemId = midiOutputSelection.getSelectedId();
        jassert(itemId != 0);        // Ensure our test has a valid comboBox item selected
        int listIndex = itemId - 1;

        auto midiOutputList = juce::MidiOutput::getAvailableDevices();

        midiOutputDevice = juce::MidiOutput::openDevice(midiOutputList[listIndex].identifier);
        jassert(midiOutputDevice);

        midiOutputDevice->startBackgroundThread();
    }


    void paint (Graphics& g) override
    {
        g.fillAll(getLookAndFeel().findColour (ResizableWindow::backgroundColourId));
    }


    void createGUI()
    {
        //
        // MIDI Inputs
        //
        auto midiInputList = juce::MidiInput::getAvailableDevices();
        int index = 1;

        for (auto x : midiInputList)
            midiInputSelection.addItem(x.name, index++);

        midiInputSelection.onChange = [this] { midiInputChanged(); };
        midiInputSelection.setSelectedId(1);

        midiInputLabel.setText("MIDI Input", NotificationType::dontSendNotification);

        addAndMakeVisible(midiInputSelection);
        addAndMakeVisible(midiInputLabel);

        //
        // MIDI Output
        //
        auto midiOutputList = juce::MidiOutput::getAvailableDevices();
        index = 1;

        for (auto x : midiOutputList)
            midiOutputSelection.addItem(x.name, index++);

        midiOutputSelection.onChange = [this] { midiOutputChanged(); };
        midiOutputSelection.setSelectedId(1);

        midiOutputLabel.setText("MIDI Output", NotificationType::dontSendNotification);

        addAndMakeVisible(midiOutputSelection);
        addAndMakeVisible(midiOutputLabel);

        //
        // BPM
        //
        
        bpm_label.setText("BPM", NotificationType::dontSendNotification);
        bpm_value.setEditable(true);
        addAndMakeVisible(bpm_label);
        addAndMakeVisible(bpm_value);
    
        bpm_value.setText(juce::String(bpm), NotificationType::dontSendNotification);

        bpm_value.onTextChange = [this]
        {
            bpm = bpm_value.getText().getIntValue();
        };

        //
        // Channel
        //
        
        channelLabel.setText("Channel", NotificationType::dontSendNotification);
        channelValue.setEditable(true);
        addAndMakeVisible(channelLabel);
        addAndMakeVisible(channelValue);
    
        channelValue.setText(juce::String(channel), NotificationType::dontSendNotification);

        channelValue.onTextChange = [this]
        {
            channel = channelValue.getText().getIntValue();
        };

        //
        // Info text
        //

        infoText.setText("Select MIDI input and output and BPM.\nHitting key on MIDI keyboard plays the last played NOTE at the given BPM.", NotificationType::dontSendNotification);
        addAndMakeVisible(infoText);
        
        setSize(400, 200);
    }


    void resized() override
    {
        midiInputSelection.setBounds(10, 30, 300, 30);
        midiInputLabel.setBounds(10, 0, 300, 30);

        midiOutputSelection.setBounds(400, 30, 300, 30);
        midiOutputLabel.setBounds(400, 0, 300, 30);

        channelLabel.setBounds(midiOutputLabel.getRight() + 10, midiOutputLabel.getY(), 80, 30);
        channelValue.setBounds(channelLabel.getX(), midiOutputSelection.getY(), 50, 30);
        channelValue.setColour(Label::ColourIds::backgroundColourId, juce::Colour(0xFF555555));

        bpm_label.setBounds(10, 100, 40, 30);
        bpm_value.setBounds(bpm_label.getRight() + 10, 100, 50, 30);
        bpm_value.setColour(Label::ColourIds::backgroundColourId, juce::Colour(0xFF555555));
        
        infoText.setBounds(10, 100, 600, 300);
    }


    void prepareToPlay(int samplesPerBlockExpected, double newSampleRate) override
    {
        sampleRate = newSampleRate;
        expectedSamplesPerAudioBlock = samplesPerBlockExpected;

        midiInputCollector.reset(newSampleRate);
    }


    void getNextAudioBlock(const AudioSourceChannelInfo& bufferToFill) override
    {
        const double currentTime = Time::getMillisecondCounter();

        // Clear audio data
        bufferToFill.clearActiveBufferRegion();

        //
        // Get MIDI input data to find out which note we want to play repeatedly
        //
        midiInputBuffer.clear();
        midiInputCollector.removeNextBlockOfMessages(midiInputBuffer, expectedSamplesPerAudioBlock);

        // Create MIDI output data
        midiOutBuffer.clear();

        for (auto x : midiInputBuffer)
        {
            auto message = x.getMessage();

            // Forward only note on/off messages
            if (!message.isNoteOnOrOff())
                continue;

            midiNoteNumber = message.getNoteNumber();
        }

        //
        // Play the note repeatedly with fixed intervals
        //
        audioSamplesPerNote = 60*sampleRate / (8*bpm);     // 32th notes
        
        if (audioSamplesToNextNote < expectedSamplesPerAudioBlock)
        {
            auto noteOffMessage = juce::MidiMessage::noteOff(channel, currentMIDINoteNumber);
            auto noteOnMessage  = juce::MidiMessage::noteOn(channel, currentMIDINoteNumber, 1.0f);
            
            // If there is an old note playing: turn it off
            if (currentMIDINoteNumber != 0)
                midiOutBuffer.addEvent(noteOffMessage, audioSamplesToNextNote);

            // If we have actually selected a note to be triggered, then play it
            if (midiNoteNumber != 0)
            {
                midiOutBuffer.addEvent(noteOnMessage, audioSamplesToNextNote);
                currentMIDINoteNumber = midiNoteNumber;
            }
        }
        
        audioSamplesToNextNote -= expectedSamplesPerAudioBlock;
        if (audioSamplesToNextNote < 0)
            audioSamplesToNextNote += audioSamplesPerNote;
        
        if (midiOutputDevice)
            midiOutputDevice->sendBlockOfMessages(midiOutBuffer, currentTime, sampleRate);
    }


    void releaseResources() override
    {
    }

private:
    double                      sampleRate = 0.0;
    int                         expectedSamplesPerAudioBlock = 0;

    juce::Label                 infoText;

    // MIDI Inputs
    juce::ComboBox              midiInputSelection;
    juce::Label                 midiInputLabel;

    juce::MidiMessageCollector  midiInputCollector;
    juce::MidiBuffer            midiInputBuffer;

    // MIDI Outputs
    juce::ComboBox              midiOutputSelection;
    juce::Label                 midiOutputLabel;

    juce::Label                 channelLabel;
    juce::Label                 channelValue;
    int                         channel = 1;
    
    juce::Label                 bpm_label;
    juce::Label                 bpm_value;
    int                         bpm = 165;

    int                         audioSamplesPerNote = 0;
    int                         audioSamplesToNextNote = 0;
    int                         midiNoteNumber = 0;
    int                         currentMIDINoteNumber = 0;
    
    std::unique_ptr<MidiOutput> midiOutputDevice;
    juce::MidiBuffer            midiOutBuffer;

    JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(AudioAppDemo)
};
